# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


class ModuleDocFragment(object):
    DOCUMENTATION = r"""
options:
    testcol2option:
        description:
            - An option taken from testcol2
        type: str
        version_added: 1.0.0
    testcol2option2:
        description:
            - Another option taken from testcol2
        type: str
"""
