from __future__ import annotations


class ModuleDocFragment:
    DOCUMENTATION = """
    options:
      fragment_expression:
        description: a fragment hosted expression that must be trusted whose default resolves to 4
        default: 2 + 2
    """
