# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: validate
short_description: validate
description: validate
author: "validate (@validate)"
"""

EXAMPLES = r"""
"""

RETURN = r"""
"""
